/*
 * TiVo initialization and setup code
 *
 * Copyright (C) 2009 TiVo Inc. All Rights Reserved.
 */
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/tivoconfig.h>
#include <linux/string.h>
#include <asm/bootinfo.h>
#include <linux/module.h>

// defined in cfe_call.c
int tivo_get_cfe_env_variable(void *name_ptr, int name_length, void *val_ptr,  int val_length);

unsigned int boardID;
EXPORT_SYMBOL(boardID);

int __init tivo_init(void)
{
    char buf[64];

    boardID = 0;
    
    // first check for brev on the command line
    char *brev_param = strstr(arcs_cmdline, "brev="); 
    if (brev_param) boardID=simple_strtoul(brev_param+5, NULL, 0)<<8;

    // if not, try the BREV variable in the flash  
    if (!boardID && !tivo_get_cfe_env_variable("BREV", 4, buf, sizeof(buf))) boardID=simple_strtoul(buf,NULL,0)<<8;

    // trap some broken Neutrons, assumes brev cannot legitimately be less than 0x1060.
    if (boardID < kTivoConfigBoardIDGen07Base)
    {
        printk("boardID = 0x%x --> 0x%x\n", boardID, kTivoConfigBoardIDNeutron);
        boardID = kTivoConfigBoardIDNeutron;
    }

    InitTivoConfig( (TivoConfigValue) boardID );
    SetTivoConfig( kTivoConfigBoardID, (TivoConfigValue) boardID );

    // Virtual hardware strapping resistors!
    unsigned int cpu=*(unsigned int *)(0xb0404000); // SUN_TOP_CTRL_PROD_REVISION
    if ((cpu >> 16) != 0x7405) boardID |= 1;        // flag non-7405s
    
    // tell user space about the boardID
    if (!strstr(arcs_cmdline, "boardID=")) 
    {
        sprintf(buf, " boardID=0x%X", boardID);
        strncat(arcs_cmdline, buf, CL_SIZE - 1 - strlen(arcs_cmdline));
    }
    
    // Use /dev/ttyS0 as the default kernel console if none specified
    if (!strstr(arcs_cmdline, "console=")) 
        strncat(arcs_cmdline, " console=ttyS0", CL_SIZE - 1 - strlen(arcs_cmdline));

    if (!strstr(arcs_cmdline, "HpkImpl=")) 
        strncat(arcs_cmdline, " HpkImpl=Gen07", CL_SIZE - 1 - strlen(arcs_cmdline));
        

    return 0;
}
